#!/usr/bin/env python3
from Bio import AlignIO
from tqdm import tqdm

import argparse
from sys import stdin, stdout, stderr
from math import log
from statistics import mean
from collections import Counter
warned = set()

class FreqCounter(Counter):
    def freqs(self):
        N = sum(self.values())
        return {b:v/N for b, v in self.most_common()}

def shannon_entropy(alncol):
    alncol = [b for b in alncol if b not in ("-", "*", "X")]
    aafreq = FreqCounter(alncol)
    entropy = -0
    for aa, freq in aafreq.freqs().items():
        entropy += freq*(log(freq,2))
    return len(alncol), -entropy

def shanent_aln(alignment, alnformat="fasta"):
    aln = AlignIO.read(alignment, alnformat)
    if len(aln) < 15 and "15_seqs" not in warned:
        print(f"WARN: {alignment} has <15 seqs, treat SE values with a grain of salt", file=stderr)
        warned.add("15_seqs")
    se = []
    for i in range(aln.get_alignment_length()):
        se.append(shannon_entropy(aln[:,i]))
    return aln, se

def main(argv=None):
    """Calculate Shannon's entropy (in bits) at each column of one or more alignments"""
    ap = argparse.ArgumentParser()
    ap.add_argument("-f", "--fofn", action="store_true",
            help="The first positional arg is a file of file names")
    ap.add_argument("-o", "--out-tsv", default=stdout, type=argparse.FileType("wt"),
            help="Output TSV file")
    ap.add_argument('alignments', nargs="+",
            help='One or more MSAs, in fasta format')
    args = ap.parse_args(argv)

    if args.fofn:
        fofn = args.alignments[0]
        if fofn == "-":
            fofn = "/dev/stdin"
        with open(fofn) as fh:
            args.alignments = [line.rstrip() for line in fh]

    print("alnfile", "nseqs",  "column", "shannon_entropy", sep="\t", file=args.out_tsv)
    for alnfile in tqdm(args.alignments, unit=" Alignment"):
        aln, se = shanent_aln(alnfile)
        for i, (l, s) in enumerate(se):
            print(alnfile, l, i+1, s, sep="\t", file=args.out_tsv)
        print(alnfile, len(aln), "*", mean([s for l, s in se]), sep="\t", file=args.out_tsv)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
import argparse

from Bio import SeqIO

def falen_main(argv=None):
    """Tabulate the lengths of sequences in a FASTA file"""
    ap = argparse.ArgumentParser()
    ap.add_argument("-t", "--total", action="store_true",
            help="Output only a total length per file")
    ap.add_argument("fastas", help="FASTA file of contigs", nargs="+")
    args = ap.parse_args(argv)
    
    for infile in args.fastas:
        with open(infile) as fh:
            file_total = 0
            for seq in SeqIO.parse(fh, 'fasta'):
                file_total += len(seq.seq)
                if not args.total:
                    print(infile, seq.name, len(seq.seq), sep="\t")
            if args.total:
                print(infile, file_total, sep="\t")
                
if __name__ == "__main__":
    falen_main()



from sys import stderr
from .pairs import main as pairs_main

def main(argv=None):
    """Add an old-style /1 /2 pair indicator to paired-end fastq files (DEPRECATED, use blsl pairs)"""
    print("WARNING: blsl pairslash is deprecated, replace use with blsl pairs", file=stderr)
    pairs_main(argv)

if __name__ == "__main__":
    main()




# Copyright (c) 2022 Dr. K. D. Murray/Gekkonid Consulting <spam@gekkonid.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import argparse
import json
from pathlib import Path
import subprocess
from base64 import b64encode 
import shutil


def get_data_uri(data):
    if isinstance(data, str) or isinstance(data, Path):
        data = Path(data)
        if data.stat().st_size > 2**22:
            raise ValueError(f"Too big to embed: {data}")
        with open(data, "rb") as fh:
            data = fh.read()
    elif len(data) > 2**22:
        raise ValueError(f"Too big to embed data")
    if data[0] == 0x1f and data[1] == 0x8b:
        mediatype = "data:application/gzip"
    else:
        mediatype = "data:application:octet-stream"

    enc_str = b64encode(data)

    data_uri = mediatype + ";base64," + str(enc_str)[2:-1]
    return data_uri


default_template = """
<!doctype html>
<html lang=en>
  <head>
    <meta charset=utf-8>
    <meta name=viewport content="width=device-width,initial-scale=1">
    <meta http-equiv=x-ua-compatible content="IE=edge,chrome=1">
    <title>__title__</title>
    <script src="https://cdn.jsdelivr.net/npm/igv@2.12.6/dist/igv.min.js"></script>
</head>
<body>
  <div id="igvherepls"></div>
  <script>
    var igvDiv = document.getElementById("igvherepls");
    var options = __data__;
        igv.createBrowser(igvDiv, options)
                .then(function (browser) {
                    console.log("Created IGV browser");
                })
  </script>
</body>
"""

def link(linkpath, target, copy=False):
    linkpath = Path(linkpath)
    target = Path(target).absolute()
    if copy:
        linkpath.parent.mkdir(exist_ok=True)
        shutil.copyfile(target, linkpath)
    else:
        try:
            if linkpath.samefile(target):
                return
        except:
            pass
        linkpath.symlink_to(target)

def genigvjs_main(argv=None):
    """Generate a simple IGV.js visualisation of some bioinf files."""
    ap = argparse.ArgumentParser()
    ap.add_argument("--template", "-T", required=False,
            help="Alternative HTML template")
    ap.add_argument("--copy", "--cp", action="store_true",
            help="Copy, don't link, data to output dir")
    ap.add_argument("--embed", "-e", action="store_true",
            help="Encode data as base64 within html (almost always a bad idea on non-toy datasets).")
    ap.add_argument("--prefix", "-p", default="./",
            help="URL prefix of paths relative to generated index.html")
    ap.add_argument("--title", "-t", default="IGV.js",
            help="Webpage title")
    ap.add_argument("--reference", "-r", required=True,
            help="reference file. must be *.fa or *.fasta, with associated *.fai, and can also have *.gff as annotation.")
    ap.add_argument("--locus", "-l", 
            help="Initial region of interest.")
    ap.add_argument("--outdir", "-o", required=True,
            help="Output directory. Data files will be softlinked there, and 'index.html' will be generated.")
    ap.add_argument("tracks", help="Any files to be used as tracks. Can be gff/bed/vcf/bcf/bam/cram. Must be indexed", nargs="+")
    args = ap.parse_args(argv)

    if args.template is not None:
        with open(args.template) as fh:
            template = fh.read()
    else:
        template = default_template

    outdir = Path(args.outdir)
    outdir.mkdir(exist_ok=True)

    data = {
        "reference": {},
        "tracks": [],
    }
    if args.locus:
        data["locus"] = args.locus
    ref = Path(args.reference)
    refbase = ref.stem
    reffai = Path(str(ref) + ".fai")
    data["reference"] = {
        "id": refbase,
        "name": refbase,
    }
    if args.embed:
        data["reference"].update({
            "fastaURL": get_data_uri(ref),
            "indexURL": get_data_uri(reffai),
        })
    else:
        data["reference"].update({
            "fastaURL": f"{args.prefix}{ref.name}",
        })
        link(outdir / ref.name, ref, copy=args.copy)
        link(outdir / (ref.name + ".fai"), reffai, copy=args.copy)
    cbpaired = ['#a6cee3','#1f78b4','#b2df8a','#33a02c','#fb9a99','#e31a1c',
                '#fdbf6f','#ff7f00','#cab2d6','#6a3d9a', '#b15928']
    for i, track in enumerate(args.tracks):
        track = Path(track)
        base = track.stem
        dots = track.name.split('.')
        format = dots[-1]
        if format == "gz":
            format = dots[-2]
        index = None
        if format == "bam":
            index = Path(str(track) + ".bai")
        else:
            index = Path(str(track) + ".tbi")
        if not index.exists():
            index = None
        trackdat = {
            "name": base,
            "format": format,
            "autoHeight": True,
            "minHeight": 50,
            "maxHeight": 500,
            "color": cbpaired[i%len(cbpaired)],
        }
        if args.embed:
            trackdat["url"] = get_data_uri(track)
            if index is not None:
                trackdat["indexURL"] = get_data_uri(index)
        else:
            trackdat["url"] =  f"{args.prefix}{track.name}"
            link(outdir / track.name, track, copy=args.copy)
            if index is not None:
                trackdat["indexURL"] =  f"{args.prefix}{index.name}"
                link(outdir / index.name, index, copy=args.copy)
        data["tracks"].append(trackdat)
    with open(outdir / "index.html", "w") as fh:
        html = template \
                .replace("__title__", args.title) \
                .replace("__data__", json.dumps(data, indent=4))
        fh.write(html)

if __name__ == "__main__":
    genigvjs_main()



# Copyright (c) 2022 Dr. K. D. Murray/Gekkonid Consulting <spam@gekkonid.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


from Bio import SeqIO
from shlex import quote
import argparse
from tqdm import tqdm


full_help = """
Repeats or other problematic sequences are often masked from a reference
genome.  Ideally, this is done by converting each base to lower-case ("soft
masking"), though is sometimes done by replacing bases with N ("hard masking").

This tool reads a genome file, and for each sequence, outputs a zero-based bed
file of the masked regions. By default this means lower-case regions, although
if -N/--hardmasked is given, stretches of N are considered masked too.
"""

def mask2bed_main(argv=None):
    """The inverse of bedtools maskfasta: softmasked fasta -> unmasked fasta + mask.bed"""
    ap = argparse.ArgumentParser(epilog=full_help)
    ap.add_argument("-N", "--hardmasked", action="store_true",
            help="count N characters as a masked base (i.e. FASTA is hard-masked)")
    ap.add_argument("fasta", help="FASTA file of contigs")
    args = ap.parse_args(argv)
    
    seqs = SeqIO.parse(args.fasta, "fasta")
    for seq in tqdm(seqs):
        last_start = 0
        last_masked = None
        for i, base in enumerate(str(seq.seq)):
            base_masked = base.islower() or (args.hardmasked and base.lower() == "n")
            if last_masked is None:
                last_masked = base_masked
                continue
            if last_masked != base_masked:
                if last_masked:
                    print(seq.name, last_start, i, sep="\t")
                else:
                    last_start = i
                last_masked = base_masked
        if last_masked:
            print(seq.name, last_start, i+1, sep="\t")

if __name__ == "__main__":
    mask2bed_main()



# Copyright (c) 2023 Dr. K. D. Murray/Gekkonid Consulting <spam@gekkonid.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import csv
import sys
import math
import argparse


def tabcat_main(argv=None):
    """Concatenate table (c/tsv) files, adding the filename as a column"""
    ap = argparse.ArgumentParser()
    ap.add_argument("-d", "--delim", default=",",
            help="Column delimiter.")
    ap.add_argument("-o", "--output", default="-", type=argparse.FileType("wt"),
            help="Column delimiter.")
    ap.add_argument("-N", "--add-name", action="store_true",
            help="Add a column with the file name")
    ap.add_argument("-f", "--fofn", action="store_true",
            help="Each input is a file of filenames")
    ap.add_argument("xsvs", help="Input {C,T}SVs to concatenate", nargs="+")
    args = ap.parse_args(argv)

    class xsv(csv.unix_dialect):
        delimiter = args.delim
        quotechar = '"'
        doublequote = True
        skipinitialspace = True
        quoting=csv.QUOTE_MINIMAL

    if args.fofn:
        files = []
        for fofn in args.xsvs:
            with open(fofn) as fh:
                for file in fh:
                    files.append(file.rstrip())
    else:
        files = args.xsvs
    outcsv = None
    for file in files:
        with open(file) as fh:
            for rec in csv.DictReader(fh, dialect=xsv):
                rec["tabcat_filename"] = file
                if outcsv is None:
                    outcsv = csv.DictWriter(args.output, dialect=xsv, fieldnames=[k for k in rec])
                    outcsv.writeheader()
                outcsv.writerow(rec)

if __name__ == "__main__":
    tabcat_main()





import argparse

def parsefai(fai):
    with open(fai) as fh:
        for l in fh:
            cname, clen, _, _, _ = l.split()
            clen = int(clen)
            yield cname, clen


def make_regions(fas, window=1e6):
    window = int(window)
    if not fas.endswith(".fai"):
        fas = fas + ".fai"
    for cname, clen in parsefai(fas):
        for start in range(0, clen, window):
            wlen = min(clen - start, window)
            yield (cname, start, start+wlen, wlen)

def main(argv):
    """Make a bed/region file of genome windows"""
    ap = argparse.ArgumentParser()
    ap.add_argument("--windowsize", "-w", default=100000, type=int,
            help="Window size of each non-overlapping region")
    ap.add_argument("--bed", "-b", type=argparse.FileType("w"),
            help="Output bed file")
    ap.add_argument("--regions", "-r", type=argparse.FileType("w"), 
            help="Output region (chr:start-stop) file")
    ap.add_argument("fasta", help="Fasta reference (must have fai)")
    args = ap.parse_args(argv)

    for chrm, start, stop, leng in make_regions(args.fasta, window=args.windowsize):
        region = f"{chrm}:{start+1}-{stop}"
        if args.bed is not None:
            print(chrm, start, stop, region, file=args.bed, sep="\t")
        if args.regions is not None:
            print(region, file=args.regions, sep="\t")



from tqdm.auto import tqdm
from collections import defaultdict
import re
from sys import stdout, stderr, stdin
from ._utils import fqparse



def pairslash(header):
    name, *comment = header.split(" ", 1)
    if name.endswith("/1") or name.endswith("/2"):
        return header
    
    i = None
    if name.endswith("/1"):
        i = 1
    if name.endswith("/2"):
        i = 2
    if comment:
        if comment[0].startswith("1:"):
            i = 1
        if comment[0].startswith("2:"):
            i = 2
    if i is None:
        raise ValueError(f"Unable to guess pair from header '{header}'")
    return f"{name}/{i} {comment[0]}"


def read_fileset(readfiles):
    res = defaultdict(dict)
    for readfile in readfiles:
        stem, read = re.match(r"(.+?)[_.]?(R[12]|SE|IL|INTERLEAVED)?(_001)?\.(FQ|FASTQ)(\.GZ)?", readfile).groups()[:2]
        read = read.upper()
        if read == "INTERLEAVED":
            read = "IL"
        if read in res[stem]:
            raise ValueError("Two files with same stem and read index: {readfiles}")
        res[stem][read] = readfile
    return dict(res)


def output_read(args, read):
    if args.pairslash:
        read[0] = pairslash(read[0])
    print("\n".join(read), file=args.output)


def handle_fileset(args, stem, fileset):
    if "R1" in fileset and "R2" in fileset:
        for r1, r2 in tqdm(zip(fqparse(fileset["R1"]), fqparse(fileset["R2"])), desc=stem):
            output_read(args, r1)
            output_read(args, r2)
    if "IL" in fileset:
        for r in tqdm(fqparse(fileset("IL")), desc=stem):
            output_read(args, r)
    if "SE" in fileset:
        for r in tqdm(fqparse(fileset("SE")), desc=stem):
            output_read(args, r)
            

def main(argv=None):
    """Handle paired-end reads, with various transformations."""
    import argparse
    ap = argparse.ArgumentParser("blsl pairs")
    ap.add_argument("--output", "-o", default=stdout, type=argparse.FileType("w"),
        help="Output file (default: stdout)")
    ap.add_argument("--pairslash", "-s", action="store_true",
        help="Add /1 /2 to paired reads")
    ap.add_argument("readfile", nargs='+', default="/dev/stdin")
    args=ap.parse_args(argv)

    for stem, fileset in read_fileset(args.readfile).items():
        handle_fileset(args, stem, fileset)

if __name__ == "__main__":
    main()




# Copyright (c) 2023 Dr. K. D. Murray/Gekkonid Consulting <spam@gekkonid.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import gzip
import sys
import argparse
from ._utils import fqparse, rc

def nstitch_main(argv=None):
    """Combine R1 + R2 into single sequences, with an N in the middle"""
    ap = argparse.ArgumentParser()
    ap.add_argument("-s", "--seq", default="N",
            help="Sequence used to glue together R1 + R2 of each pair")
    ap.add_argument("-q", "--qual", default="!",
            help="Quality score for glue sequences")
    ap.add_argument("-I", "--input-interleaved", action="store_true",
            help="Only read R1 which contains interleaved sequences")
    ap.add_argument("r1", help="R1 fastq.gz")
    ap.add_argument("r2", help="R2 fastq.gz", nargs="?")
    args = ap.parse_args(argv)
    
    if args.input_interleaved:
        r1 = gzip.open(args.r1, "rt")
        seqs = zip(fqparse(r1), fqparse(r1))
    else:
        r1 = gzip.open(args.r1, "rt")
        r2 = gzip.open(args.r2, "rt")
        seqs = zip(fqparse(r1), fqparse(r2))
    for s1, s2 in seqs:
        n1, n2 = s1[0].lstrip("@"), s2[0].lstrip("@")
        i1 = n1.split(" ", 1)[0]
        i2 = n2.split(" ", 1)[0]
        if i1 != i2:
            print(f"ERROR: read IDs don't match: {i1} {i2}", file=sys.stderr)
            sys.exit(1)
        newseq = f"{s1[1]}{args.seq}{rc(s2[1])}"
        newqual = f"{s1[3]}{args.qual*len(args.seq)}{s2[3]}"
        assert len(s1[1]) == len(s1[3])
        assert len(s2[1]) == len(s2[3])
        assert len(newseq) == len(newqual)
        print(f"@{n1} {n2}")
        print(newseq)
        print("+")
        print(newqual)
    r1.close()
    r2.close()

if __name__ == "__main__":
    nstitch_main()



#!/usr/bin/env python3
import argparse
from sys import stdout
from .gffparse import parseGFF3, write_line
from tqdm import tqdm

def main(argv=None):
    """Sanitise a messy gff attribute column to just simple tags """
    ap = argparse.ArgumentParser("gfftagsane")
    ap.add_argument("-o", "--output", type=argparse.FileType("wt"), default=stdout,
            help="Output gff file")
    ap.add_argument("-f", "--fields", default="ID,Name,Parent,locus_tag",
            help="Attribute tags to keep (case sensitive, give multiple times like -f ID -f tag2 -f tag3).")
    ap.add_argument("input")

    args = ap.parse_args(argv)
    args.fields = args.fields.split(",")

    for record in tqdm(parseGFF3(args.input, return_as=dict)):
        attrs = {k: record["attributes"][k] for k in args.fields if k in record["attributes"]}
        record["attributes"] = attrs
        write_line(record, args.output)

if __name__ == "__main__":
    main()



# Copyright (c) 2022 Dr. K. D. Murray/Gekkonid Consulting <spam@gekkonid.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from Bio import SeqIO
from shlex import quote
import argparse

def calc_n50(fastafile):
    lengths = []
    seqs = SeqIO.parse(fastafile, "fasta")
    for seq in seqs:
        lengths.append(len(seq.seq))
    lengths.sort()
    asmlen = sum(lengths)
    cumsum = 0
    for l in lengths:
        cumsum += l
        if cumsum >= asmlen/2:
            return sum(lengths), len(lengths), l

def n50_main(argv=None):
    """Calculate N50 and total length of a set of contigs"""
    ap = argparse.ArgumentParser()
    ap.add_argument("fastas", help="FASTA file of contigs", nargs="+")
    args = ap.parse_args(argv)
    
    for infile in args.fastas:
        nbases, ncontig, n50 = calc_n50(infile)
        print(quote(infile), ":", sep="")
        print("  ncontigs:", ncontig)
        print("  n50:", n50)
        print("  total_length:", nbases)


if __name__ == "__main__":
    n50_main()



from io import BytesIO
from pathlib import Path
from sys import stdin, stdout, stderr
from tqdm import tqdm
from collections import Counter
import random
import gzip

from ._utils import fqparse

def main(argv=None):
    """Sample a fraction of read pairs from an interleaved fastq file
    """
    import argparse
    ap = argparse.ArgumentParser()
    ap.add_argument("-o", "--out", default="/dev/fd/1", type=argparse.FileType("wb"),
            help="Output file")
    ap.add_argument("-f", "--frac", default=0.1, type=float,
            help='Fraction of reads to sample')
    ap.add_argument("-i", "--input", type=Path, default="/dev/fd/0",
            help="Input fastq file")
    ap.add_argument("-s", "--singleend", action="store_const", const=4, default=8, dest="nlines",
            help="Single-end reads (default: paired end, interleaved)")
    args=ap.parse_args(argv)
    
    if args.input.suffix == ".gz":
        stream = gzip.open(args.input, "rb")
    else:
        stream = open(args.input, "rb")

    total = 0
    sampled = 0
    for pair in tqdm(fqparse(stream, args.nlines), unit="read pairs"):
        total += 1
        if random.random() < args.frac:
            sampled += 1
            args.out.write(b"".join(pair))

    print(f"Done! Sampled {sampled} of {total} pairs", file=stderr)

if __name__ == "__main__":
    main()



import gzip
import os

def rc(seq):
    d = {"A": "T", "G":"C", "C":"G", "T":"A"}
    return "".join([d.get(x, "N") for x in reversed(list(seq.upper()))])

def fqparse(stream, n=4):
    if isinstance(stream, os.PathLike) or isinstance(stream, str):
        if str(stream).endswith(".gz"):
            with gzip.open(stream, "rt") as fh:
                yield from fqparse(fh, n)
        else:
            with open(stream, "rt") as fh:
                yield from fqparse(fh, n)
    else:
        fqp = list()
        for line in stream:
            fqp.append(line.rstrip("\n"))
            if len(fqp) == n:
                yield fqp
                fqp = list()
        if len(fqp) == n:
            yield fqp
        assert len(fqp) == 0




#!/usr/bin/env python3
from blsl.gffparse import parseGFF3, write_gene, gffInfoFields, write_gff_line
import argparse


def main(argv=None):
    """Format a reasonably compliant GFF for use with bcftools csq"""
    ap = argparse.ArgumentParser()
    ap.add_argument("-o", "--output", default="/dev/stdout",
            help="Output GFF compatible with bcftools csq")
    ap.add_argument("input", help="Input GFF")
    args = ap.parse_args(argv)

    with open(args.output, "w") as fh:
        for entry in parseGFF3(args.input):
            if entry["type"] == "gene":
                id = entry["attributes"]["ID"]
                entry["attributes"]["ID"] = f"gene:{id}"
                entry["attributes"]["biotype"] =  "protein_coding"  # FIXME properly parse this
            elif entry["type"] in ("mRNA", ):
                id = entry["attributes"]["ID"]
                parent = entry["attributes"]["Parent"]
                entry["attributes"]["ID"] = f"transcript:{id}"
                entry["attributes"]["Parent"] = f"gene:{parent}"
                entry["attributes"]["biotype"] =  "protein_coding"  # FIXME properly parse this
            elif entry["type"] in ("exon", "CDS", "five_prime_UTR", "three_prime_UTR"):
                parent = entry["attributes"]["Parent"]
                entry["attributes"]["Parent"] = f"transcript:{parent}"
                #if entry["type"] == "CDS" and entry["attributes"]["original_annotator"] == "Liftoff" and entry["phase"] is None:
                #    entry["phase"] = 0
            else:
                continue
            write_gff_line(entry, file=fh)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
from sys import stdin, stdout, stderr, exit, argv
import re
import argparse

def main(argv):
    """Make a ncbi-style acc2taxid.map file for a uniref fasta"""

    prog = "blsl uniref-acc2taxid"
    ap = argparse.ArgumentParser(prog=prog)
    ap.add_argument("--strip-uniref", "-s", action="store_true",
            help="Strip UniRefXX_ prefixes from IDs (for Diamond, see https://github.com/bbuchfink/diamond/issues/639)")
    args = ap.parse_args(argv)
    
    if stdout.isatty() and stdin.isatty():
        print(f"USAGE: zcat uniref.fasta.gz | {prog} [-s] >uniref.acc2taxid.tsv")
        exit(0)

    #if args.strip_uniref:
    #    print("Stripping UniRefXX_ from each sequence ID (for Diamond)", file=stderr)
    print("accession.version", "taxid", sep="\t", file=stdout)
    for line in stdin:
        if not line.startswith(">"):
            continue
        seqid, *line = line.rstrip().lstrip(">").split(" ")
        fields = dict(x.split("=", 1) for x in line if "=" in x)
        if args.strip_uniref:
            # needed for Diamond, see https://github.com/bbuchfink/diamond/issues/639
            seqid = re.sub(r"UniRef\d\d_", "", seqid)
        print(seqid, fields.get("TaxID", 0), sep="\t", file=stdout)

if __name__ == "__main__":
    main(argv)



# Copyright (c) 2022 Dr. K. D. Murray/Gekkonid Consulting <spam@gekkonid.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from Bio import SeqIO
import re
import argparse
from collections import Counter

def telogrep_main(argv=None):
    """Search contigs for known telomere repeats"""
    ap = argparse.ArgumentParser()
    ap.add_argument("fasta", help="FASTA file of contigs")
    ap.add_argument(
            "-w", "--window", type=int, default=500, metavar="BASES",
            help="Scan BASES (default 500) at the end of each contig for telomeres",
    )
    ap.add_argument(
            "-t", "--threshold", type=float, default=30, metavar="N",
            help="If >N% of the window is telomeric repeats, consider that a telomere",
    )
    ap.add_argument(
            "-o", "--omit-telomereless", action="store_true",
            help="Don't output for contigs with neither 5' or 3' telomeres"
    )
    ap.add_argument(
            "-T", "--tsv", action="store_true",
            help="Output as a TSV for easy reading into other tools."
    )
    args = ap.parse_args(argv)

    # These are taken from https://dx.doi.org/10.1093/gbe/evt019
    tel_forward = "C{2,4}T{1,2}A{1,3}"
    tel_reverse = "T{1,3}A{1,2}G{2,4}"

    window = args.window
    min_bp = window * args.threshold/100
    nrev = 0
    nfwd = 0
    nt2t = 0

    if args.tsv:
        print("contig", "contig_len", "n_bp_fwd", "teloseq_fwd", "n_bp_rev", "teloseq_rev", sep="\t")

    for record in SeqIO.parse(args.fasta, "fasta"):
        ctgid = str(record.description).split(" ")[0]
        sequence = str(record.seq).upper()

        fwds = re.findall(tel_forward, sequence[:window])
        fwdlens = [len(x) for x in fwds]
        fwdbp = sum(fwdlens)

        revs = re.findall(tel_reverse, sequence[-window:])
        revlens = [len(x) for x in revs]
        revbp = sum(revlens)

        n_t = 0
        fwdstr = ""
        mcfwd = ""
        if fwdbp > min_bp:
            mcfwd = Counter(fwds).most_common(1)[0][0]
            fwdstr = "{}bp of ({})".format(fwdbp, mcfwd)
            nfwd += 1
            n_t += 1
        else:
            fwdbp = 0

        revstr = ""
        mcrev = ""
        if revbp > min_bp:
            mcrev = Counter(revs).most_common(1)[0][0]
            revstr = "{}bp of ({})".format(revbp, mcrev)
            nrev += 1
            n_t += 1
        else:
            revbp = 0
        
        if n_t == 2:
            nt2t += 1

        if args.omit_telomereless and fwdstr == "" and revstr == "":
            continue
        
        if args.tsv:
            print(ctgid, len(sequence), fwdbp, mcfwd, revbp, mcrev, sep="\t")
        else:
            print('{:<20} ({:>9}bp) {:>20} ------- {:<20}'.format(ctgid, len(sequence), fwdstr, revstr))
    
    if not args.tsv:
        print()
        print('{:<20}               {:>20} ------- {:<20}  ({} T2T chroms)'.format(
                "TOTAL:", "{} 5' Telos".format(nfwd), "{} 3' Telos".format(nrev), nt2t))

if __name__ == "__main__":
    telogrep_main()



#!/usr/bin/env python3
from tqdm import tqdm
from cyvcf2 import VCF

import sys
from sys import stdin, stdout, stderr
from subprocess import Popen, PIPE
import argparse
import math
import json
from concurrent.futures import ProcessPoolExecutor, as_completed

def parallel_regions(vcf, cores=1):
    V = VCF(vcf)
    # 10 chunks per chrom per core
    chunks = len(V.seqlens)*100*cores
    for cname, clen in zip(V.seqnames, V.seqlens):
        chunk = 10000#int(math.ceil(clen/chunks))
        for start in range(0, clen, chunk):
            s = start+1
            e = start+chunk+1
            yield f"{cname}:{s}-{e}"

def variant2dict(v, fields=None):
    for i, alt in enumerate(v.ALT):
        dat = {"CHROM": v.CHROM, "POS": v.POS, "REF": v.REF, "ALT": alt, "QUAL": v.QUAL}
        dat["call_rate"] = v.call_rate
        for key, val  in v.INFO:
            if isinstance(val, str) and "," in val:
                val = val.split(',')
            if isinstance(val, tuple) or isinstance(val, list):
                val = val[i]
            dat[f"INFO_{key}"] = val
        if fields:
            dat = {K:V for K, V in dat.items() if K in fields}
        yield json.dumps(dat)

def one_chunk_stats(vcf, chunk, fill=True, fields=None):
    cmd=f"bcftools view -r {chunk} {vcf} -Ou"
    if fill:
        cmd = f"{cmd} | bcftools +fill-tags - -Ou -- -d -t all,F_MISSING"
    res = []
    with Popen(cmd, shell=True, stdout=PIPE) as proc:
        vcf = VCF(proc.stdout)
        for v in vcf:
            for r in variant2dict(v, fields=fields):
                res.append(r)
        del vcf
    return res

def chunkwise_bcfools_stats(args):
    regions = list(parallel_regions(args.vcf))
    with tqdm(total=len(regions), unit="chunk") as pbar:
        for i in range(0, len(regions), 10000):
            to=min(len(regions), i+10000)
            with ProcessPoolExecutor(args.threads) as exc:
                jobs = (exc.submit(one_chunk_stats, args.vcf, region, fill=args.fill_tags_first, fields=args.fields) for region in regions[i:to])
                for job in as_completed(jobs):
                    pbar.update(1)
                    for res in job.result():
                        args.output.write(res)
                        args.output.write("\n")


def main(argv=None):
    """Use bcftools to calculate various statistics, outputing an R-ready table"""
    ap = argparse.ArgumentParser()
    ap.add_argument("-o", "--output", default=stdout, type=argparse.FileType("wt"),
            help="Output jsonlines file")
    ap.add_argument("-f", "--fields", nargs="+",
            help="Use only these fields")
    ap.add_argument("-F", "--fill-tags-first", action="store_true",
            help="Use bcftools +fill-tags to pre-fill more fields (forces no parallelism)")
    ap.add_argument("-t", "--threads", default=1, type=int,
            help="Number of parallel threads")

    ap.add_argument("vcf")
    args = ap.parse_args(argv)

    chunkwise_bcfools_stats(args)


if __name__ == "__main__":
    main()



#!/usr/bin/python3
from shutil import get_terminal_size
from collections import Counter
from sys import stdin, stdout, stderr, exit
from math import ceil, log10
import select

def main(argv=None):
    """Make a summary histogram of git-annex-list output"""
    if not select.select([stdin], [], [], 1)[0]:
        print("USAGE:  git annex list ... | blsl galhist", file=stderr)
        print("This tool makes a histogram of file occurence patterns to summarise the output of git annex list. Just pipe git anex list in on stdin.", file=stderr)
        exit(1)

    twidth = 80
    try:
        if stdout.isatty():
            twidth = get_terminal_size()[0]
    except:
        pass
        
    header = []
    counts = Counter()
    for line in stdin:
        if not header:
            header.append(line.rstrip())
        elif line.startswith("|"):
            header.append(line.rstrip())
        else:
            key, file = line.split(" ", maxsplit=1)
            counts[key]+=1 

    for l in header:
        print(l)
    _, mcv = counts.most_common(1)[0]
    nw = int(ceil(log10(mcv)))
    remaining = twidth - len(header) - 1 - nw - 2 

    for k, v in counts.most_common():
        vbarw = min(round(v/mcv*remaining), remaining)
        vbar = ("#" * vbarw).ljust(remaining, " ")
        npad = str(v).rjust(nw, " ")
        print(f"{k} {vbar} ({npad})")



#!/usr/bin/env python3
try:
    from tqdm import tqdm
except ImportError:
    def tqdm(*args, **kwargs):
        return args[0]

import argparse
import re
from pathlib import Path
from sys import stderr, exit


def make_replacer(mode, reffa=None, org_name=None, delim="~", outdelim="~", refdelim="~", lowercase=True):
    def L(string):
        if lowercase:
            return string.lower()
        return string
    names = []
    if not reffa.endswith(".fai"):
        reffa += ".fai"
    with open(reffa) as fh:
        for line in fh:
            names.append(line.split()[0])
    fromto = {}
    for name in names:
        if org_name is not None:
            splitname = [org_name, "0", name]
        else:
            splitname = name.strip().split(refdelim)
        pansn_name = delim.join(splitname)
        if mode == "cat": 
            fromto[pansn_name] = L(outdelim.join(splitname))
        elif mode == "add":
            fromto[splitname[2]] = L(outdelim.join(splitname))
        elif mode == "rm":
            fromto[pansn_name] = L(splitname[2])
    return fromto


def do_tsv(infh, outfh, cols, replacer, coldelim="\t", skipcomment=False):
    for line in tqdm(infh):
        if line.startswith("#"):
            if not skipcomment:
                outfh.write(line)
            continue
        fields = line.rstrip("\r\n").split(coldelim)
        for col in cols:
            fields[col] = replacer[fields[col]]
        print(*fields, sep=coldelim, file=outfh)


def do_fasta(infh, outfh, replacer):
    for line in infh:
        if not line.startswith(">"):
            outfh.write(line)
            continue
        name = line.lstrip(">").rstrip("\n\r").split(" ")[0]
        name = replacer[name]
        print(">", name, sep="", file=outfh)


fasta_exts = {".fa", ".fna", ".fasta"}
tsv_exts = {".gff", ".gff3", ".gtf", ".bed", ".bg", ".tsv"}


def main(argv=None):
    """Add, remove, or modify PanSN-style prefixes to contig/chromosome names in references"""
    extra="""
    Modes:

        cat: pass through, ensuring all names are PanSN-style. Useful to convert delimiters in PanSN names.

        rm: remove the PanSN specific prefix, e.g. at9900_1_chr1 -> chr1
        
        add: add a PanSN-specific prefix, e.g. chr1 -> at9900_1_chr1.
    """
    ap = argparse.ArgumentParser("blsl pansn-rename", epilog=extra)
    ap.add_argument("-C", "--colsep", default="\t",
        help="Delimiter between columns in c/tsv file. (default tab)")
    ap.add_argument("-d", "--delim", default="_",
        help="Delimiter between fields in PanSN names in <INPUT> (the # in indiv#1#chr1)")
    ap.add_argument("-l", "--lowercase", action="store_true",
        help="Lowercase all names")
    ap.add_argument("-D", "--out-delim",
        help="--delim for the output. Defaults to the same as --delim.")
    ap.add_argument("-r", "--reference-fasta", required=True,
        help="Reference fasta file with PanSN names.")
    ap.add_argument("-R", "--ref-delim",
        help="Delimiter between fields in PanSN names in the --reference-fasta file (default: same as --delim)")
    ap.add_argument("-n", "--org-name",
        help="Organism name (or similar) to prepend to current identifiers (separated by --out-delim).")
    ap.add_argument("-c", "--colidx", action="append", type=int,
        help="Which column in TSV should be sub'd? (default depends on input "
             "filetype, normally 0, use more than once for multiple columns "
             "e.g. -c 0 -c 3 for 1st & 4th cols).")
    ap.add_argument("-m", "--mode", choices=["cat", "rm", "add"],
        help="What should I do? (see below)")
    ap.add_argument("-o", "--output", default="-", type=argparse.FileType("w"),
        help="output filename, default: stdout")
    ap.add_argument("input", type=argparse.FileType("r"))
    args = ap.parse_args(argv)

    if args.reference_fasta is None and args.org_name is None:
        print("ERROR: must give either --reference-file or --org-name as source of new names")
        exit(1)
    if args.colidx is None:
        args.colidx = [0, ]
    args.colidx = list(sorted(set(args.colidx)))
    if args.out_delim is None:
        args.out_delim = args.delim
    if args.ref_delim is None:
        args.ref_delim = args.delim
        
    replacer = make_replacer(args.mode, args.reference_fasta, args.org_name, args.delim, args.out_delim, args.ref_delim, lowercase=args.lowercase)
    
    input_ftype = "tsv"
    try:
        input_filename = Path(args.input.name)
        if input_filename.suffix in fasta_exts:
            input_ftype = "fasta"
        elif input_filename.suffix in tsv_exts:
            input_ftype =  "tsv"
        else:
            pass
            #print("WARNING: unknown filetype", input_filename.suffix, ", assuming TSV-like format", file=stderr)
    except:
        print("WARNING: unable to determine input filetype, assuming TSV-like format", file=stderr)
    
    if input_ftype == "tsv":
        do_tsv(args.input, args.output, args.colidx, replacer, coldelim=args.colsep)
    else:
        do_fasta(args.input, args.output, replacer)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
from Bio import Entrez
from tqdm import tqdm

import argparse
from time import sleep
from sys import stdout, stdin, stderr, exit
from pathlib import Path
import io

def esearch(*args, chunks=10000, **kwargs):
    kwargs["retmax"] = chunks
    ids = []
    n = None
    i = 0
    with tqdm(desc="esearch IDs") as progress:
        while n is None or i < n:
            this = Entrez.esearch(*args, **kwargs, retstart=i)
            for x in range(3):
                try:
                    sleep(0.5)
                    this = Entrez.read(this)
                    break
                except RuntimeError:
                    sleep(0.5)
                    pass
            else:
                this = Entrez.read(this)
            nret = len(this["IdList"])
            ids.extend(this["IdList"])
            i += nret
            n = int(this["Count"])
            progress.total = n
            progress.update(nret)
    return ids
            

def efetch(db="nucleotide", rettype="genbank", retmode="text", ids=None, chunks=2000):
    with tqdm(desc=f"efetch {rettype}", total=len(ids)) as progress:
        for start in range(0, len(ids), chunks):
            stop = min(start + chunks, len(ids))
            idstr = ",".join(ids[start:stop])
            res = Entrez.efetch(db=db, retmode=retmode, rettype=rettype, id=idstr)
            if start != 0 and rettype=="runinfo":
                # Skip header on non-first one
                hdr = next(res)
            for line in res:
                if isinstance(line, bytes):
                    line=line.decode('utf-8')
                yield line
            progress.update(stop-start)


def main(argv=None):
    """Use the Entrez API to search for and download something. A CLI companion to the NCBI search box"""
    ap = argparse.ArgumentParser()
    ap.add_argument("-e", "--email", type=str,
            help="Email for Entrez API")
    ap.add_argument("-c", "--chunksize", type=int, default=1000,
            help="Number of records to download per chunk")
    ap.add_argument("-o", "--out", type=argparse.FileType('w'), default="-",
            help="Genbank output")
    ap.add_argument("-d", "--db", required=True,
            help="db api param")
    ap.add_argument("-r", "--rettype", required=True,
            help="rettype api param")
    ap.add_argument("-m", "--retmode", default="text",
            help="API return mode")
    ap.add_argument("-i", "--id-file", type=Path,
            help="List of IDs (give with --term to save a list of IDs to this file, or without to use this list of IDs)")
    ap.add_argument("-t", "--term", type=str)
    args = ap.parse_args(argv)

    Entrez.email = args.email

    if args.term:
        ids = esearch(term=args.term, db=args.db)
        if args.id_file:
            with args.id_file.open("w") as fh:
                for id in ids:
                    print(id, file=fh)
    elif args.id_file:
        ids = []
        with args.id_file.open("r") as fh:
            for line in fh:
                ids.append(line.rstrip())
    else:
        print("ERROR: must give either --term to search for, or a list of ids in --id-file", file=stderr)
        exit(1)

    for line in efetch(ids=ids, db=args.db, rettype=args.rettype, retmode=args.retmode, chunks=1000):
        args.out.write(line)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright 2020 K.D. Murray
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


import re
from Bio import SeqIO
from hashlib import md5
import json


class FastaSanitiser(object):
    def __init__(self, regex=r'(S\d+)', forgiving=True, usemd5=False, old_idmap=None):
        self.regex = regex
        if isinstance(regex, str):
            self.regex = re.compile(regex)
        self.id_map = {}
        self.forgiving = forgiving
        self.usemd5 = usemd5

        if isinstance(old_idmap, str):
            with open(old_idmap) as fh:
                self.id_map = json.load(fh)
        elif isinstance(old_idmap, dict):
            self.id_map = old_idmap


    def _get_id(self, seqid):
        if self.usemd5:
            return md5(seqid.encode("utf-8")).hexdigest()
        m = self.regex.search(seqid)
        if m is not None:
            return m.group(0)
        elif m is None and self.forgiving:
            return md5(seqid.encode("utf-8")).hexdigest()
        else:
            raise ValueError(f"Couldn't match regex to sanitise name '{seqid}'")

    def sanitise(self, seqfile, format='fasta'):
        seqs = []
        with open(seqfile) as fh:
            for seq in SeqIO.parse(seqfile, format):
                print(seq.name, seq.id, seq.description)
                oldname = seq.description  # stupidly, some "ids" have spaces in them in gisaid data so we use the whole line.
                newname = self._get_id(oldname)
                if newname in self.id_map:
                    raise ValueError(f"Duplicate sequences with sanitised id {newname}")
                self.id_map[newname] = oldname
                seq.name = newname
                seq.id = newname
                seq.description = ""
                seqs.append(seq)
        return seqs

    def convert_forward(self, source, destination):
        conv_seqs = self.sanitise(source)
        with open(destination, "w") as seqfh, open(destination + ".idmap.json", "w") as mapfh:
            SeqIO.write(conv_seqs, seqfh, "fasta")
            json.dump(self.id_map, mapfh)

    def convert_back(self, source, destination):
        with open(source) as ifh, open(destination, "w") as ofh:
            for line in ifh:
                for id_new, id_old in self.id_map.items():
                    if id_new in line:
                        line = line.replace(id_new, id_old)
                        print(line, end="")
                ofh.write(line)


def main(argv=None):
    """Sanitise fasta IDs to something sane, then back again"""
    import argparse
    ap = argparse.ArgumentParser()
    ap.add_argument("--backwards", "-b", action="store_true",
                    help="Convert from sanitised to unsantised format")
    ap.add_argument("--old-mapping", "-m", type=str,
                    help="Json file mapping clean IDs to old IDs.")
    ap.add_argument("--regex", "-r", type=str, default=r'(EPI_ISL_\d+)',
                    help="ID regex. Must capture the new ID as the first group.")
    ap.add_argument("--forgiving", "-f", action='store_true',
                    help="If --regex isn't found, use md5sum of ID instead.")
    ap.add_argument("--md5sum", action='store_true',
                    help="Instead of using --regex, always use md5sum of SeqIDs.")
    ap.add_argument("--output", "-o", type=str,
                    help="Output file")
    ap.add_argument("input", type=str,
                    help="Input file")
    args = ap.parse_args(argv)
    fs = FastaSanitiser(regex=args.regex, forgiving=args.forgiving, usemd5=args.md5sum, old_idmap=args.old_mapping)
    if not args.backwards:
        fs.convert_forward(args.input, args.output)
    else:
        fs.convert_back(args.input, args.output)


if __name__ == "__main__":
    main()



# Copyright (c) 2023 Dr. K. D. Murray/Gekkonid Consulting <spam@gekkonid.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import csv
import sys
import math
import argparse
from collections import defaultdict

def recursive_defaultdict():
    x = defaultdict(recursive_defaultdict)
    x["__count__"] = 0
    return x

def recursive_taxtable(lineage, level=0):
    yield (level, lineage["__name__"], lineage["__count__"], lineage["__line__"])
    for k, v in reversed(sorted(filter(lambda x: x[0] not in ["__count__", "__name__", "__line__"],
                                       lineage.items()),
                                key=lambda x: x[1]["__count__"])):
        yield from recursive_taxtable(v, level=level+1)

def insert_lineage(tree, lin, max_level=None, n=1):
    X = tree
    X["__count__"] += n
    X["__name__"] = "root"
    X["__line__"] = "NA"
    for i, t in enumerate(lin):
        X = X[t]
        X["__count__"] += n
        X["__name__"] = t
        X["__line__"] = ";".join(lin)
        if max_level is not None and i >= max_level:
            break

def gg2k_main(argv=None):
    """Summarise a table with GreenGenes-style lineages into a kraken-style report."""
    ap = argparse.ArgumentParser()
    ap.add_argument("-l", "--lca-column", type=lambda x: int(x) - 1,
            help="Take the lowest common ancestor per value in this column")
    ap.add_argument("-f", "--column", required=True, type=lambda x: int(x) - 1,
            help="Which column contains lineages?")
    ap.add_argument("-d", "--column-delim", default="\t",
            help="Quality score for glue sequences")
    ap.add_argument("-D", "--lineage-delim", default=";",
            help="Delimiter between nodes in a taxon lineage")
    ap.add_argument("-L", "--max-level", default=None, type=lambda x: int(x) - 1,
            help="Maximum level to descend to.")
    ap.add_argument("-V", "--vsearch-size", default=None, type=lambda x: int(x) - 1,
            help="Parse vsearch size=NNNN headers from this column, use NNNN as a count rather than # hits/reads.")
    ap.add_argument("table", help="Table of taxon hits")
    args = ap.parse_args(argv)
    
    
    class OurDialect(csv.unix_dialect):
        delimiter = args.column_delim
        quotechar = '"'
        doublequote = True
        skipinitialspace = True

    lineage = recursive_defaultdict()
    with open(args.table) as fh:
        current_id = None
        lca = []
        for rec in csv.reader(fh, dialect=OurDialect):
            linstr = rec[args.column]
            lin = linstr.split(args.lineage_delim)
            n = 1
            if args.lca_column is not None:
                if current_id != rec[args.lca_column]:
                    if args.vsearch_size is not None and current_id is not None:
                        hdr = current_id
                        fields = dict(x.split("=") for x in hdr.split(";") if "=" in x)
                        n = int(fields["size"])
                    insert_lineage(lineage, lca, args.max_level, n=n)
                    lca = lin
                    current_id = rec[args.lca_column]
                for i, tax in enumerate(lin):
                    if i >= len(lca) or lca[i] != tax:
                        break
                lca = lca[:i+1]
            else:
                if args.vsearch_size is not None:
                    hdr = rec[args.vsearch_size]
                    fields = dict(x.split("=") for x in hdr.split(";") if "=" in x)
                    n = int(fields["size"])
                insert_lineage(lineage, lin, args.max_level, n=n)
        if lca:
            if args.vsearch_size is not None:
                hdr = current_id
                fields = dict(x.split("=") for x in hdr.split(";") if "=" in x)
                n = int(fields["size"])
            insert_lineage(lineage, lca, args.max_level, n=n)

    n = int(math.ceil(math.log10(lineage["__count__"]+1)))
    taxtbl = list(recursive_taxtable(lineage))
    max_name = max(len(t[1]) for t in taxtbl)
    max_level = max(t[0] for t in taxtbl)
    for level, name, count, linstr in taxtbl:
        linstr = linstr.split(args.lineage_delim, level)
        linstr = args.lineage_delim.join(linstr[:-1])
        print(name.ljust(max_name + (max_level-level)*2).rjust(max_name + max_level * 2), str(count).rjust(n), level, linstr, sep="|")

if __name__ == "__main__":
    gg2k_main()




# Copyright (c) 2023 Dr. K. D. Murray/Gekkonid Consulting <spam@gekkonid.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import csv
import sys
import math
import argparse

class blast6(csv.unix_dialect):
    delimiter = "\t"
    quotechar = '"'
    doublequote = False
    escapechar = "\\"
    skipinitialspace = True

def output_best(recs, fields):
    nbegin = len(recs)
    mineval =None
    maxalnlen=None
    if nbegin < 1:
        return
    mineval = min(float(x["evalue"]) for x in recs)
    recs = [
        rec for rec in recs if float(rec["evalue"]) <= mineval
    ]
    maxalnlen = max(float(x["length"]) for x in recs)
    recs = [
        rec for rec in recs if float(rec["length"]) >= maxalnlen
    ]
    nend=len(recs)
    for rec in recs:
        print(*[rec[x] for x in fields], sep="\t")
    

def equalbestblast_main(argv=None):
    """Output only the best blast hits."""
    ap = argparse.ArgumentParser()
    ap.add_argument("-f", "--outfmt", default="qseqid sseqid pident length mismatch gapopen qstart qend sstart send evalue bitscore",
            help="Blast outfmt headers (default is the default for blast -outfmt 6)")
    ap.add_argument("table", help="Table of blast hits")
    args = ap.parse_args(argv)
    
    fields = args.outfmt.split(" ")
    query = None
    hits = []
    with open(args.table) as fh:
        for rec in csv.DictReader(fh, dialect=blast6, fieldnames=fields):
            if query != rec["qseqid"]:
                if query is not None:
                    output_best(hits, fields)
                hits = []
                query = rec["qseqid"]
            hits.append(rec)
        if query is not None:
            output_best(hits, fields)

if __name__ == "__main__":
    equalbestblast_main()




# Copyright (c) 2022 Dr. K. D. Murray/Gekkonid Consulting <spam@gekkonid.com> 
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from sys import argv, exit, stderr

__version__ = "0.3.0"

cmds = {}


from .telogrep import telogrep_main
cmds["telogrep"] = telogrep_main

from .n50 import n50_main
cmds["n50"] = n50_main

from .falen import falen_main
cmds["falen"] = falen_main

from .mask2bed import mask2bed_main
cmds["mask2bed"] = mask2bed_main

from .pansn_rename import main as pansn_rename_main
cmds["pansn-rename"] = pansn_rename_main

from .genigvjs import genigvjs_main
cmds["genigvjs"] = genigvjs_main

from .ildemux import main as ildemux_main
cmds["ildemux"] = ildemux_main

from .ilsample import main as ilsample_main
cmds["ilsample"] = ilsample_main

from .regionbed import main as regionbed_main
cmds["regionbed"] = regionbed_main

from .uniref_accessionmap import main as uniref_main
cmds["uniref-acc2taxid"] = uniref_main

from .nstitch import nstitch_main
cmds["nstitch"] = nstitch_main

from .gg2k import gg2k_main
cmds["gg2k"] = gg2k_main

from .equalbestblast import equalbestblast_main
cmds["equalbestblast"] = equalbestblast_main

from .tabcat import tabcat_main
cmds["tabcat"] = tabcat_main

from .esearchandfetch import main as esf_main
cmds["esearchandfetch"] = esf_main

from .deepclust2fa import deepclust2fa_main
cmds["deepclust2fa"] = deepclust2fa_main

from .farename import farename_main
cmds["farename"] = farename_main

from .gffcat import gffcat_main
cmds["gffcat"] = gffcat_main

from .gffparse import gffparse_main
cmds["gffparse"] = gffparse_main

from .gffcsqify import main as gffcsqify_main
cmds["gffcsqify"] = gffcsqify_main

from .gfftagsane import main as gfftagsane_main
cmds["gfftagsane"] = gfftagsane_main

from .liftoff_gff3 import liftoff_gff3_main
cmds["liftoff-gff3"] = liftoff_gff3_main

from .ebiosra2rl2s import main as rl2s_main
cmds["ebiosra2rl2s"] = rl2s_main

from .galhist import main as galhist_main
cmds["galhist"] = galhist_main

from .pairslash import main as pairslash_main
cmds["pairslash"] = pairslash_main

from .pairs import main as pairs_main
cmds["pairs"] = pairs_main


try:
    from .vcfstats import main as vcfstats_main
    cmds["vcfstats"] = vcfstats_main

    from .vcfparallel import main as vcfparallel_main
    cmds["vcfparallel"] = vcfparallel_main
except ImportError as exc:
    if len(argv) < 2:
        print(str(exc), "-- disabling vcfstats command", file=stderr)

from .shannon import main as shannon_main
cmds["shannon-entropy"] = shannon_main

from .fastasanitiser import main as fastasanitiser_main
cmds["fastasanitiser"] = fastasanitiser_main

from .tidyqc import main as tidyqc_main
cmds["tidyqc"] = tidyqc_main

from .jsonl2csv import main as jsonl2csv_main
cmds["jsonl2csv"] = jsonl2csv_main


def mainhelp(argv=None):
    """Print this help message"""
    print("USAGE: blsl <subtool> [options...]\n\n")
    print("Where <subtool> is one of:\n")
    for tool, func in cmds.items():
        print("  {:<19}".format(tool + ":"), " ", func.__doc__.split("\n")[0])
    print("\n\nUse blsl subtool --help to get help about a specific tool")

cmds["help"] = mainhelp

def main():
    if len(argv) < 2:
        mainhelp()
        exit(0)
    if argv[1] not in cmds:
        print("ERROR:", argv[1], "is not a known subtool. See help below")
        mainhelp()
        exit(1)
    cmds[argv[1]](argv[2:])





from tqdm import tqdm

import re
import argparse
from sys import stdin, stdout, stderr
from pathlib import Path
from collections import Counter, defaultdict
import json


def as_rtype(rtype, val):
    try:
        return rtype(val)
    except ValueError:
        pass
    return None

class PatternMatcher:
    name = "Base Class"
    patterns = {}
    def match(self, file):
        if hasattr(self, "firstline"):
            with open(file) as fh:
                line = next(fh)
                return re.match(self.firstline, line) is not None
        if hasattr(self, "anyline"):
            with open(file) as fh:
                for line in fh:
                    if re.match(self.anyline, line) is not None:
                        return True
                return False

    
    def read_patterns(self, file):
        def extract(line):
            for name, (rtype, regex) in self.patterns.items():
                m = re.match(regex, line)
                if m is not None:
                    return name, as_rtype(rtype, m[1])
            return None, None
        res = {}
        with open(file) as fh:
            for line in fh:
                key, val = extract(line)
                if key is not None:
                    res[key] = val
        return res

    def read(self, file):
        res = {
                "filename": str(file),
                "tool": self.name,
        }
        if hasattr(self, "patterns"):
            res.update(self.read_patterns(file))
        return res


class RSEMLog(PatternMatcher):
    name="RSEM"
    anyline = r"\d+ reads; of these:"
    patterns = {
            "total_reads": (int, r"(\d+) reads; of these:"),
            "paired_reads": (int, r"\s*(\d+) \(.+%\) were paired; of these:"),
            "unaligned_reads": (int, r"\s*(\d+) \(.+%\) aligned concordantly 0 times"),
            "mapped_reads": (int, r"\s*(\d+) \(.+%\) aligned concordantly exactly 1 time"),
            "multimapped_reads": (int, r"\s*(\d+) \(.+%\) aligned concordantly >1 times"),
            "overall_alignment_rate": (float, r"\s*([\d.]+)% overall alignment rate"),
    }


class AdapterRemoval(PatternMatcher):
    name="AdapterRemoval"
    anyline = r"AdapterRemoval ver. .+"
    patterns = {
            "total_reads": (int, r"Total number of read pairs: (\d+)"),
            "reads_with_adapter": (int, r"Number of reads with adapters[1]: (\d+)"),
            "reads_kept": (int, r"Number of retained reads: (\d+)"),
            "bases_kept": (int, r"Number of retained nucleotides: (\d+)"),
            "avergage_length_postqc": (float, r"Average length of retained reads: ([\d.]+)"),
    }



def main(argv=None):
    """What if MultiQC was in the tidyverse? (and much worse)"""
    tools = {RSEMLog(), AdapterRemoval()}
    ap = argparse.ArgumentParser()
    ap.add_argument("-o", "--out", type=argparse.FileType("wt"), default=stdout,
            help="Output file (jsonl)")
    ap.add_argument("files", nargs="+", type=Path)

    args = ap.parse_args(argv)

    results = defaultdict(list)
    toolcounts = Counter()

    for infile in tqdm(args.files):
        for proc in tools:
            if proc.match(infile):
                results[proc.name].append(proc.read(infile))
                toolcounts[proc.name] += 1

    for tool in results:
        n = toolcounts[tool]
        print(f"Found {n} {tool} files", file=stderr)
        for res in results[tool]:
            json.dump(res, args.out)
            args.out.write("\n")

if __name__ == "__main__":
    main()



# Copyright (c) 2023 Dr. K. D. Murray/Gekkonid Consulting <spam@gekkonid.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import csv
import sys
import argparse
from collections import defaultdict
from pathlib import Path
import gzip

from Bio.SeqIO import parse
from tqdm import tqdm


class tsv(csv.unix_dialect):
    delimiter = "\t"
    quotechar = '"'
    doublequote = False
    escapechar = "\\"
    skipinitialspace = True


def deepclust2fa_main(argv=None):
    """Split a .faa by the clusters diamond deepclust finds"""
    ap = argparse.ArgumentParser()
    ap.add_argument("-z", "--min-cluster-size", type=int, default=1,
            help="Minimum cluster size")
    ap.add_argument("-b", "--buffer", action="store_true",
            help="Buffer sequences rather than writing (more ram, more speed)")
    ap.add_argument("-c", "--clusters", required=True,
            help="Diamond deepclust result")
    ap.add_argument("--orthofinder", "--of",  action="store_true",
            help="--clusters is actually an orthofinder result, like OG: seq1 seq2 seq3")
    ap.add_argument("-f", "--faa", required=True,
            help="Fasta sequences (that you gave to diamond deepclust --db)")
    ap.add_argument("-o", "--outdir", required=True, type=Path,
            help="Output directory (files named by their centroids)")
    args = ap.parse_args(argv)

    seq2cent = {}
    cent2seq = defaultdict(list)
    with open(args.clusters) as fh:
        if args.orthofinder:
            for line in fh:
                og, seqs = line.split(":")
                og = og.strip()
                seqs = seqs.split(", ")
                for seq in seqs:
                    seq = seq.strip()
                    seq2cent[seq] = og
                    cent2seq[og].append(seq)
        else:
            for rec in csv.reader(fh, dialect=tsv):
                if rec[0] == "centroid" and rec[1] == "member":
                    continue
                c, s = rec[0:2]
                seq2cent[s] = c
                cent2seq[c].append(s)
    cent2seq_new = {}
    seq2cent_new = {}
    for cent, seqs in cent2seq.items():
        if len(seqs) >= args.min_cluster_size:
            cent2seq_new[cent] = seqs
            for seq in seqs:
                seq2cent_new[seq] = cent
    cent2seq = cent2seq_new
    seq2cent = seq2cent_new
    print(f"Read clusters, {len(cent2seq)} clusters with size >= {args.min_cluster_size}", file=sys.stderr)

    for cent in cent2seq:
        outf = args.outdir/ (c+".fa")
        with open(outf, "w") as fh:
            pass

    if args.faa.endswith(".gz"):
        fh = gzip.open(args.faa, "rt")
    else:
        fh = open(args.faa)

    buffers = defaultdict(list)
    for seq in tqdm(parse(fh, "fasta")):
        centroid = seq2cent.get(seq.id)
        if centroid is None:
            continue
        if args.buffer:
            buffers[centroid].append((seq.id, seq.seq))
        else:
            outf = args.outdir/ (centroid+".fa")
            with open(outf, "a") as ofh:
                print(f">{seq.id}", file=ofh)
                print(seq.seq, file=ofh)
    fh.close()

    for centroid, seqs in buffers.items():
        outf = args.outdir/ (centroid+".fa")
        with open(outf, "w") as ofh:
            for seq in seqs:
                print(f">{seq[0]}", file=ofh)
                print(seq[1], file=ofh)


if __name__ == "__main__":
    deepclust2fa_main()




#!/usr/bin/env python3
"""
A simple parser for the GFF3 format.

Test with transcripts.gff3 from
http://www.broadinstitute.org/annotation/gebo/help/gff3.html.

Format specification source:
http://www.sequenceontology.org/gff3.shtml

Version 1.1: Python3 ready
Version 2.0: @kdm9's version. Add code for comparison & normalisation
"""
import argparse
from collections import namedtuple, Counter
import gzip
import urllib.request, urllib.parse, urllib.error
from sys import stderr, stdout
from copy import deepcopy
import re

from tqdm.auto import tqdm

__author__  = "Uli Köhler"
__license__ = "Apache License v2.0"
__version__ = "2.1"

#Initialized GeneInfo named tuple. Note: namedtuple is immutable
gffInfoFields = ["seqid", "source", "type", "start", "end", "score", "strand", "phase", "attributes"]
GFFRecord = namedtuple("GFFRecord", gffInfoFields)

ontological_order = [
    "chromosome", "gene", "pseudogene", "pseudogenic_region", "mRNA", "transcript", "pseudogenic_transcript",
    "intron", "exon", "pseudogenic_exon", "five_prime_UTR", "three_prime_UTR", "CDS", "pseudogenic_CDS",
]

def parseGFFAttributes(attributeString, lax=False):
    """Parse the GFF3 attribute column and return a dict"""#
    if attributeString == ".": return {}
    ret = {}
    if lax:
        # The idea here is to split not just on ;, but on whole ;key=XXXXXX
        # sections. This allows e.g. unescaped ; etc in the attrs column
        attrs = re.findall(r"(^|(?<=;))([\w_-]+)=(.+?)((?=;[\w_-]+=)|$)", attributeString)
        attrs = [(x[1], x[2]) for x in attrs]
    else:
        attrs = [x.partition("=") for x in attributeString.split(";")]
        attrs = [(x[0], x[2]) for x in attrs]
    for key, value in attrs:
        if not key and not value:
            continue
        if not value:
            value = "True"
        ret[urllib.parse.unquote(key)] = urllib.parse.unquote(value)
    return ret

def parseGFF3(filename, return_as=dict, lax=False):
    """
    A minimalistic GFF3 format parser.
    Yields objects that contain info about a single GFF3 feature.
    
    Supports transparent gzip decompression.
    """
    #Parse with transparent decompression
    openFunc = gzip.open if str(filename).endswith(".gz") else open
    with openFunc(filename, "rt") as infile:
        for line in infile:
            #if line.startswith("###"):
            #    ### Yield last gene if we move that here
            #    pass
            if line.startswith("#") or line.strip() == "":
                continue
            parts = line.strip().split("\t")
            #If this fails, the file format is not standard-compatible
            assert len(parts) == len(gffInfoFields)
            #Normalize data
            normalizedInfo = {
                "seqid": None if parts[0] == "." else urllib.parse.unquote(parts[0]),
                "source": None if parts[1] == "." else urllib.parse.unquote(parts[1]),
                "type": None if parts[2] == "." else urllib.parse.unquote(parts[2]),
                "start": None if parts[3] == "." else int(parts[3]),
                "end": None if parts[4] == "." else int(parts[4]),
                "score": None if parts[5] == "." else float(parts[5]),
                "strand": None if parts[6] == "." else urllib.parse.unquote(parts[6]),
                "phase": None if parts[7] == "." else urllib.parse.unquote(parts[7]),
                "attributes": parseGFFAttributes(parts[8], lax=True)
            }
            #Alternatively, you can emit the dictionary here, if you need mutability:
            #    yield normalizedInfo
            yield return_as(**normalizedInfo)
            

def gff_heirarchy(filename, progress=None, make_missing_genes=False):
    level_canonicaliser = {
        "transcript": "mRNA",
    }
    levels = {
        "repeat_region": "l1",
        "pseudogene": "l1",
        "pseudogenic_region": "l1",
        "transposable_element_gene": "l1",
        "gene": "l1",
        "tRNA": "l2",
        "pseudogenic_transcript": "l2",
        "tmRNA": "l2",
        "mRNA": "l2",
        "exon": "l3",
        "intron": "l3",
        "CDS": "l3",
        "five_prime_UTR": "l3",
        "three_prime_UTR": "l3",
    }
    ignore = {
        "source",
        "stop_codon",
        "start_codon",
    }
    records = {}
    l2l1 = {}
    last = {"l1": None, "l2": None, "l3": None}
    i = 0
    warned = set()
    recordsrc =  parseGFF3(filename, return_as=dict)
    if progress is not None:
        if progress == True:
            desc=None
        else:
            desc=progress
        recordsrc = tqdm(recordsrc, desc=desc)
    for i, record in enumerate(recordsrc):
        typ = record["type"]
        typ = level_canonicaliser.get(typ, typ)
        record["type"] = typ
        lvl = None
        if "RNA" in typ.upper():
            lvl = "l2"
        lvl = levels.get(typ, lvl)
        if lvl is None:
            if typ not in ignore:
                print(f"WARNING: {typ} is not a nice feature, skipping", file=stderr)
                ignore.add(typ)
            continue
        try:
            id = record["attributes"]["ID"]
        except:
            if "fakeid" not in warned:
                print(f"WARNING: no id for {record}, faked to fakeid{i}")
                warned.add("fakeid")
            id = f"fakeid{i}"
        if lvl == "l1":
            record["children"] = {}
            records[id] = record
            i = 0
        elif lvl == "l2":
            try:
                parent = record["attributes"]["Parent"]
            except KeyError:
                if make_missing_genes:
                    parent = f"fakegenefor_{id}"
                    record["attributes"]["Parent"] = parent
                    records[parent] = deepcopy(record)
                    records[parent]["type"] = "gene"
                    records[parent]["attributes"] = {"ID": parent}
                    records[parent]["children"] = {}
                else:
                    parent = None
            l2l1[id] = parent
            record["children"] = {}
            try:
                records[parent]["children"][id] = record
            except KeyError:
                if "missing_l2" not in warned:
                    print(f"L2 entry {id} parent {parent} not in records? {record}")
                    warned.add("missing_l2")
        else:
            try:
                parent = record["attributes"].get("Parent", record["attributes"].get("transcript_id", None))
                if parent in records and not parent in l2l1 and make_missing_genes:
                    # We have L1 -> L3 only
                    if "made_l2" not in warned:
                        print("Made a fake mRNA for L2-less gene")
                        warned.add("made_l2")
                    new_parent = f"fakemrna_{parent}"
                    if new_parent not in l2l1:
                        new_mrna = deepcopy(records[parent])
                        new_mrna["type"] = "mRNA"
                        new_mrna["attributes"] = {"ID": new_parent, "Parent": parent}
                        new_mrna["children"] = {}
                        record["attributes"]["Parent"] = new_parent
                        l2l1[new_parent] = records[parent]
                        records[parent]["children"][new_parent] = new_mrna
                    top = parent
                    parent = new_parent
                    record["attributes"]["Parent"] = parent
                else:
                    top = l2l1[parent]
                if id in records[top]["children"][parent]["children"]:
                    i += 1
                    id = f"{id}_{i}"
                    record["attributes"]["ID"] = id
                records[top]["children"][parent]["children"][id] = record
            except KeyError:
                if "missing_l3" not in warned:
                    print(f"L3 entry {id} parent '{parent}' not in GFF?")
                    warned.add("missing_l3")
    return records


def attr2line(attrs):
    from urllib.parse import quote as Q
    return ";".join("=".join((Q(k), Q(v))) for k, v in attrs.items())


def reformat_names(gene, geneid=None, changenames=True):
    def prefix_name(entry, sid):
        name = entry["attributes"].get("Name")
        if name:
            name = f"{sid}_{name}"
        else:
            name = sid
        entry["attributes"]["Name"] = name
    if not geneid:
        geneid = gene["attributes"]["ID"]
    gene["attributes"]["ID"] = geneid
    subids = Counter()
    for child in gene.get("children", {}).values():
        ct = child["type"]
        subids[ct] += 1
        ci = subids[ct]
        subid = f"{geneid}_{ct}{ci:02d}"
        child["attributes"]["ID"] = subid
        child["attributes"]["Parent"] = geneid
        if changenames:
            prefix_name(child, subid)
        #print("E", child["attributes"]["Name"])
        gcids = Counter()
        for gchild in child.get("children", {}).values():
            gt = gchild["type"]
            gcids[gt] += 1
            gi = gcids[gt]
            gcid = f"{geneid}_{gt}{gi:02d}"
            #print("F", gt, gcid)
            gchild["attributes"]["ID"] = gcid
            gchild["attributes"]["Parent"] = subid
            #print("G", gcid, gchild["attributes"].get("Name"))
            if changenames:
                prefix_name(gchild, gcid)
            #print("H", gcid, gchild["attributes"]["Name"])
    

def write_gff_line(line, file=None):
    attr = ";".join(f"{k}={urllib.parse.quote(v)}" for k, v in line["attributes"].items())
    cols = [line.get(f, ".") for f in gffInfoFields[:-1]] + [attr]
    cols = ["." if x is None else x for x in cols]
    print(*cols, file=file, sep="\t")

def write_line(entry, file):
    x = [entry[field] if entry[field] is not None else "." for i, field in enumerate(gffInfoFields)]
    x[-1] = attr2line(x[-1])
    print(*x, sep="\t", file=file)

def write_gene(gene, geneid=None, file=None, changenames=False):
    if geneid or changenames:
        reformat_names(gene, geneid, changenames)
    write_line(gene, file)
    for child in gene.get("children", {}).values():
        write_line(child, file)
        for gchild in child.get("children", {}).values():
            write_line(gchild, file)       
    print("###", file=file)
    

def get_all_features(annot, ftype):
    feat = set()
    if annot["type"] == ftype:
        feat.add((annot["start"], annot["end"], annot["strand"], annot["type"]))
    for child in annot.get("children", {}).values():
        feat.update(get_all_features(child, ftype))
    return feat


def features_equal(a, b, ftype):
    a = set(get_all_features(a, ftype))
    b = set(get_all_features(b, ftype))
    return a == b


def feature_distance(a, b, ftype):
    def overlap(a, b):
        return a[0] <= b[0] <= a[1] or a[0] <= b[1] <= a[1]
    a = list(get_all_features(a, ftype))
    b = list(get_all_features(b, ftype))
    ovl = list()
    noovl = list()
    b_with_ovl = set()
    for ai in range(len(a)):
        for bi in range(len(b)):
            if overlap(a[ai], b[bi]):
                ovl.append((a[ai], b[bi]))
                b_with_ovl.add(b[bi])
                break
        else:
            # a does not overlap with an b
            noovl.append(a[ai])
    for bi in range(len(b)):
        if b[bi] not in b_with_ovl:
            noovl.append(b[bi])
    dist = 0
    for A, B in ovl:
        dist += abs(A[1] - B[1]) + abs(A[0] - B[0])
    for x in noovl:
        dist += x[1] - x[0]
    return dist


def gffparse_main(argv=None):
    """Format a GFF sanely"""
    ap = argparse.ArgumentParser()
    ap.add_argument("-o", "--output", default="/dev/stdout",
            help="Output file")
    ap.add_argument("-t", "--tabular", action="store_true",
            help="Convert gff including attributes to simple TSV")
    ap.add_argument("-c", "--prefix-chrom", action="store_true",
            help="Prefix gene names with chromsome name (e.g. useful for concatenating augustus results)")
    ap.add_argument("-g", "--make-missing-genes", action="store_true",
            help="Invent Fake L1 'gene' entries for any L2 lacking an L1 parent")
    ap.add_argument("input", help="Input GFF")
    args = ap.parse_args(argv)

    with open(args.output, "w") as fh:
        if args.tabular:
            records =  list(tqdm(parseGFF3(args.input, return_as=dict), desc="Parse GFF", item="line"))
            fields = set()
            for line in records:
                fields.update(line["attributes"].keys())
            print(*gffInfoFields[:-1], *fields, sep="\t", file=fh)
            for line in tqdm(records, desc="Tabularise"):
                cols = [line[x] for x in gffInfoFields[:-1]] + [line["attributes"].get(x) for x in fields]
                cols = [x if x is not None else "" for x in cols]
                print(*cols, sep="\t", file=fh)
        else:
            gff = gff_heirarchy(args.input, progress="Parse   ", make_missing_genes=args.make_missing_genes)
            for _, gene in tqdm(gff.items(), desc="Process "):
                if args.prefix_chrom:
                    newgid = f"{gene['seqid'].replace('#', '_')}_{gene['attributes']['ID']}"
                    reformat_names(gene, geneid=newgid, changenames=False)
                write_gene(gene, file=fh)

if __name__ == "__main__":
    gffparse_main()



#!/usr/bin/env python3
import argparse
from collections import OrderedDict, defaultdict
from sys import stdin, stdout, stderr

reserved_keys = ["ID", "Name", "Alias", "Parent", "Target", "Gap",
        "Derives_from", "Note", "Dbxref", "Ontology_term", "Is_circular"]

def fixcase(key):
    if key not in reserved_keys:
        return key.lower()
    else:
        return key

def liftoff_gff3_main(argv=None):
    """Obtain an actually-useful GFF3 from Liftoff by fixing basic GFF3 format errors"""
    ap = argparse.ArgumentParser("blsl liftoff-gff3")
    ap.add_argument("-o", "--output", type=argparse.FileType("w"), default="-",
            help="Output mostly-kosher GFF3 file")
    ap.add_argument("input", type=argparse.FileType("r"), default='-',
            help="Input liftoff 'gff' file.")
    args=ap.parse_args(argv)
    
    emptyattrcount = 0
    ids = defaultdict(set)
    replacement_suffixes = dict()
    replacements = dict()
    for lineno, line in enumerate(args.input):
        if lineno == 0 and not line.startswith("##gff-version"):
            print("##gff-version 3", file=args.output)
        if line.startswith("#"):
            args.output.write(line)
            continue
        fields = line.rstrip("\n").split("\t")
        if len(fields) != 9:
            print("WARNING: wrong number of columns on line", lineno, f"\n{line}", file=stderr)
        type = fields[2]
        kvp = OrderedDict([(fixcase(x[0]),x[1]) for x in [y.split("=") for y in fields[-1].split(";")]])
        if "" in kvp:  # sometime you get ID=***;=; type bullshit in there, clear it out
            emptyattrcount += 1
            del kvp[""]
        if "Parent" in kvp:
            parents = kvp["Parent"].split(",")
            for i, parent in enumerate(parents):
                if parent in replacements:
                    parents[i] = replacements[parent]
            parents = [p for p in parents if not p.endswith("-Protein")]  # remove protein parents as the aren't defined anywhere
            kvp["Parent"] = ",".join(parents)
        if "ID" in kvp:
            if type.lower() == "gene" and kvp["ID"] in ids[type] :
                # Many data types are *supposed* to have the same ID for a
                # block of discontinuous entries. We handle genes like this,
                # and then just add this number again to the exon/mRNA/CDS.
                oldname = kvp["ID"]
                newname = oldname
                for i in range(2, 9999999999):
                    if newname not in ids[type]:
                        break
                    newname = f"{oldname}_{i}"
                    replacement_suffixes[newname] = i
                ids[type].add(newname)
                kvp["ID"] = newname
                kvp["kdm_liftoff_fix_oldid"] = oldname
                print("WARNING: duplicate", type, "ID:", oldname, f"at {fields[0]}:{fields[3]}-{fields[4]} now called", newname, file=stderr)
                replacements[oldname] = newname
            elif "Parent" in kvp and kvp["Parent"] in replacement_suffixes:
                oldname = kvp["ID"]
                suf = replacement_suffixes[kvp['Parent']]
                newname = f"{oldname}_{suf}"
                kvp["ID"] = newname
                kvp["kdm_liftoff_fix_oldid"] = oldname
                print("WARNING: duplicate", type, "ID:", oldname, f"at {fields[0]}:{fields[3]}-{fields[4]} now called", newname, file=stderr)
                replacements[oldname] = newname
                replacement_suffixes[newname] = suf
                ids[type].add(kvp["ID"])
            else:
                ids[type].add(kvp["ID"])
        kvglue = ";".join(f"{k}={v}" for k, v in kvp.items())
        fields[-1] = kvglue
        print(*fields, sep="\t", file=args.output)
    if emptyattrcount > 0:
        print("Fixed empty attribute (;=;)", emptyattrcount, "times", file=stderr)



from io import BytesIO
from pathlib import Path
import sys
from sys import stdin
from tqdm import tqdm
from collections import Counter

from ._utils import fqparse

def hamming_cloud(seq, distance=1):
    """Generate DNA seqs whose Hamming distance from seq is <= distance

    >>> sorted(hamming_cloud('GAT', 0))
    ['GAT']
    >>> sorted(hamming_cloud('GAT', 1))
    ['AAT', 'CAT', 'GAA', 'GAC', 'GAG', 'GAN', 'GCT', 'GGT', 'GNT', 'GTT', 'NAT', 'TAT']
    >>> max(hammdist(x, "GATTACA") for x in hamming_cloud('GAT', 2))
    2
    """
    from itertools import combinations, product
    seqs = set([seq, ])
    if distance == 0:
        return seqs
    elif distance > 0:
        seqs.update(hamming_cloud(seq, distance=distance-1))
    ACGTN = "ACGTN"
    for idx in combinations(range(len(seq)), distance):
        for rep in product(range(len(ACGTN) - 1), repeat=distance):
            mutated = list(seq)
            for p, r in zip(idx, rep):
                if mutated[p] == ACGTN[r]:
                    # we have N-1 replacements (hence -1 above), so we only
                    # cacluated the product above for N-1. So when we find the
                    # "replacement" that keeps the base the same, it's actually
                    # the replacement by the final letter in the alphabet.
                    r = -1
                mutated[p] = ACGTN[r]
            seqs.add(''.join(mutated))
    return seqs


class ByteBucket(object):
    """
    This bullshit is needed to avoid having too many files open (linux
    typically limits this to 1024). This thing buffers lines and only opens one
    file at a time to write output.

    And yes it needs pigz installed, if you're reading this that's why it
    crashed.
    """
    threads = 32
    ziplevel = 6
    def biff2pigz(self, somebytes, outfile):
        from subprocess import Popen, PIPE
        with Popen(["pigz", "-n", "-p", str(self.threads), f"-{self.ziplevel}"], stdin=PIPE, stdout=outfile, bufsize=0) as proc:
            proc.stdin.write(somebytes)

    def __init__(self, outfile, size=2**27, sname=""):
        self.outfile = str(outfile)
        self.buf = BytesIO()
        self.maxsize = size
        self.first = True
        self.sname = sname

    def writelines(self, lines):
        self.write("\n".join(lines).encode('ascii'))
        self.write(b"\n")

    def write(self, somebytes):
        if self.first:
            # test write
            with open(self.outfile, "wb"):
                pass
            self.first = False
        if isinstance(somebytes, str):
            somebytes.encode('ascii')
        self.buf.write(somebytes)

        if len(self.buf.getbuffer()) > self.maxsize:
            self.flush()

    def flush(self):
        if len(self.buf.getbuffer()) > 0:
            #assert(self.first)
            mode = "wb" if self.first else "ab"
            self.first=False
            with open(self.outfile, mode, buffering=0) as fh:
                if self.outfile.endswith(".gz"):
                    self.biff2pigz(self.buf.getbuffer(), fh)
                else:
                    fh.write(self.buf.getbuffer())
                self.buf = BytesIO()

    def __del__(self):
        self.flush()

def gethdrdat(hdr):
    spot, idx = hdr.rstrip("\n").split(" ")
    idxpair = idx.split(":")[3]
    return spot, idx, idxpair

def fqp2idx(fqp):
    """fastq pair to index pair ACGT+ACGT"""
    if len(fqp) == 4:
        s1, _, ip1 = gethdrdat(fqp[0])
        return ip1
    else:
        assert(len(fqp)==8)
        s1, _, ip1 = gethdrdat(fqp[0])
        s2, _, ip2 = gethdrdat(fqp[4])
        assert(s1==s2 and ip1==ip2)
        return ip1

def make_sample_map(tablefile, outdir, fileext=".fq", distance=1):
    from csv import DictReader
    from itertools import product
    smap = {}
    with open(tablefile) as tabfh:
        for sample in DictReader(tabfh, dialect="excel-tab"):
            sname = sample["sample"]
            opath = outdir / f"{sname}.{fileext}"
            bucket = ByteBucket(opath, sname=sname)
            for i7, i5 in product(hamming_cloud(sample["i7"], distance=distance),
                                  hamming_cloud(sample["i5"], distance=distance)):
                i7plusi5 = f"{i7}+{i5}"
                if i7plusi5 in smap:
                    print("ERROR: duplicate i7+i5. reduce -m or cry into your beer.")
                    sys.exit(1)
                smap[i7plusi5] = bucket
    return smap


def main(argv=None):
    """Demultiplex modern illumina reads from read headers.

    Supports *ONLY* out-of-read indicies (in @header lines e.g. 1:N:0:ACGT+TGCA), and only reads with i7 + i5.
    """

    import argparse
    ap = argparse.ArgumentParser()
    ap.add_argument("-o", "--outdir", required=True, type=Path,
            help="Output directory (must exist)")
    ap.add_argument("-c", "--justcount", action="store_true",
            help="Write nothing, only tablulate (largely useless)")
    ap.add_argument("-g", "--guess", action="store_true",
            help="Write nothing, tablulate all indicies found (don't need a keyfile)")
    ap.add_argument("-z", "--zip", nargs="?", default=None, const=6, type=int,
            help="Compress outputs with pigz (must be installed).")
    ap.add_argument("-j", "--threads", default=8,
            help="Number of compression threads")
    ap.add_argument("-m", "--mismatch", default=1, type=int,
            help="Number tolerable mismatches (hamming distance).")
    ap.add_argument("-s", "--singleend", action="store_const", const=4, default=8, dest="nlines",
            help="Single-end reads (default: paired end, interleaved)")
    ap.add_argument("-k", "--keyfile",
            help="Mapping of i7/i5 -> sample as tsv (with cols i7, i5, sample)")
    args=ap.parse_args(argv)

    ByteBucket.threads = args.threads
    if args.zip is not None:
        ByteBucket.ziplevel = args.zip

    if args.guess:
        args.justcount=True

    fileext = "il.fastq.gz" if args.zip is not None else "il.fastq"

    if not args.guess:
        samps = make_sample_map(args.keyfile, args.outdir, fileext=fileext, distance=args.mismatch)
        print("set up sample map with", len(samps), "mappings of barcodes to files")
    
        file_undef = ByteBucket(args.outdir/f"undefined.{fileext}", sname="undefined")
        file_undef.sname = "undefined"

    stats = Counter()
    try:
        for pair in tqdm(fqparse(stdin, n=args.nlines)):
            idxpair = fqp2idx(pair)
            if args.guess:
                stats[idxpair] += 1
                continue
            ofile = samps.get(idxpair, file_undef)
            stats[ofile.sname] += 1
            if not args.justcount:
                ofile.writelines(pair)
    finally:
        with open(args.outdir / "stats.tsv", "w") as fh:
            print("sample", "nspots", file=fh, sep="\t")
            for samp, n in stats.most_common():
                print(samp, n, file=fh, sep="\t")

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from tqdm import tqdm
from cyvcf2 import VCF

from sys import stdin, stdout, stderr
import shutil
import subprocess
import argparse
import tempfile
from concurrent.futures import ProcessPoolExecutor, as_completed
from pathlib import Path
import math
import uuid
import re


def parallel_regions(vcf, cores=1, npercore=10):
    V = VCF(vcf)
    # 10 chunks per chrom per core
    chunks = len(V.seqlens)*npercore*cores
    for cname, clen in zip(V.seqnames, V.seqlens):
        chunk = int(max(min(clen, 1000), math.ceil(clen/chunks)))
        for start in range(0, clen, chunk):
            s = start+1
            e = min(clen, start+chunk+1)
            yield f"{cname}:{s:09d}-{e:09d}"


def one_chunk(vcf, chunk, temp_prefix, filterarg="", pipeline="", index=True):
    ofile = f"{temp_prefix}{chunk}.bcf"
    cmd=f"bcftools view -r {chunk} {filterarg} -Ou {vcf}"
    if pipeline:
        cmd = f"{cmd} | {pipeline}"
    if index:
        cmd = f"{cmd} | bcftools view -Ob0 --write-index -o {ofile}"
    cmd = f"({cmd}) >{ofile} 2>{ofile}.log"
    subprocess.run(cmd, shell=True, check=True)
    return ofile


def chunkwise_pipeline(args):
    filestomerge = []
    with ProcessPoolExecutor(args.threads) as exc:
        jobs = []
        if args.regions is None:
            regions = parallel_regions(args.vcf)
        else:
            regions = set()
            for line in args.regions:
                chrom, start0, end, *_ = line.rstrip("\n").split("\t")
                start0 = int(start0)
                regions.add(f"{chrom}:{start0+1}-{end}")
            regions = list(sorted(regions))
        for region in regions:
            jobs.append(exc.submit(one_chunk, args.vcf, region, args.temp_prefix, filterarg=args.filter, pipeline=args.commands, index=not args.merge_with_cat))
        for job in tqdm(as_completed(jobs), total=len(jobs), unit="chunk", desc="chunkwise variant pipeline"):
            ofile = job.result()
            filestomerge.append(ofile)
    return filestomerge


def merge_one(files, prefix, threads=1, merge_type="fast"):
    fofn = f"{prefix}fofn.txt"
    output = f"{prefix}output.bcf"
    with open(fofn, "w") as fh:
        for file in sorted(files):
            print(file, file=fh)
    merge = "--allow-overlaps --rm-dup all" if merge_type == "slow" else ""
    cmd = f"(bcftools concat --file-list {fofn} {merge} --threads {threads} -Ob0 --write-index -o {output}) >{output}.log 2>&1"
    try:
        subprocess.run(cmd, shell=True, check=True)
    except subprocess.CalledProcessError:
        with open(output + ".log") as fh:
            print(fh.read())
        raise
    return output


def merge_results(args, filestomerge):
    if args.merge_with_cat:
        with open(args.output, "wb") as ofh:
            for file in tqdm(filestomerge, desc="merge with cat", unit="file"):
                with open(file, "rb") as fh:
                    shutil.copyfileobj(fh, ofh)
        return args.output
    filestomerge =list(filestomerge)
    if len(filestomerge) > 1000:
        n_groups = int(math.ceil(math.sqrt(len(filestomerge))))
        print(f"Grouped merge -> {n_groups} groups")
        groupsize = int(math.ceil(len(filestomerge) / n_groups))
        groups = [filestomerge[i:min(i+groupsize, len(filestomerge))] for i in range(0, len(filestomerge), groupsize)]
        final_merge = []
        with ProcessPoolExecutor(args.threads) as exc:
            jobs = []
            for i, files in enumerate(groups):
                jobs.append(exc.submit(merge_one, files, f"{args.temp_prefix}merge_group_{i}_", args.merge_type))
            for job in tqdm(as_completed(jobs), total=len(jobs), unit="group"):
                ofile = job.result()
                final_merge.append(ofile)
    else:
        final_merge = filestomerge

    fofn = f"{args.temp_prefix}final_fofn.txt"
    with open(fofn, "w") as fh:
        for file in sorted(final_merge):
            print(file, file=fh)
    index = "--write-index" if re.match(r"[zb]", args.outformat) else ""
    merge = "--allow-overlaps --rm-dup all" if args.merge_type == "slow" else ""
    print(f"Finally merging {len(final_merge)} files to {args.output}")
    cmd = f"bcftools concat --file-list {fofn} {merge} --threads {args.threads} -O{args.outformat} -o {args.output} {index} >{args.output}.log 2>&1"
    subprocess.run(cmd, shell=True, check=True)
    return args.output


def main(argv=None):
    """Parallelise a bcf processing pipeline across regions"""
    ap = argparse.ArgumentParser()
    ap.add_argument("-R", "--regions", type=argparse.FileType("rt"),
            help="Use regions defined in this BED file. NB! Will only do these regions and will not check for overlaps. Be careful. Cols should be chrom, start0, end, ... and tab separated (i.e. a BED file).")
    ap.add_argument("-o", "--output", required=True,
            help="Output V/BCF file")
    ap.add_argument("-O", "--outformat", default="z8",
            help="Output file format (passed directly to bcftools -O, see man bcftools but z=vcf.gz and b=bcf)")
    ap.add_argument("-T", "--temp-prefix", default=None,
            help="Temporary output directory")
    ap.add_argument("-S", "--slow-merge", default="fast", action="store_const", const="slow", dest="merge_type",
            help="Use slow bcftools merging (with --allow-overlaps and --remove-duplicates)")
    ap.add_argument("-f", "--filter", default="", type=str,
            help="bcftools view arguments for variant filtering")
    ap.add_argument("-c", "--commands", default="", type=str,
            help="command(s) to operate. Must take uncompressed bcf on stdin and yield bcf (i.e -Ob0) on stdout. Can use | and other shell features.")
    ap.add_argument("-M", "--merge-with-cat", action="store_true",
            help="command(s) produce a non-BCF file, so merge with 'cat'.")
    ap.add_argument("-t", "--threads", default=1, type=int,
            help="Number of parallel threads")
    ap.add_argument("vcf")
    args = ap.parse_args(argv)

    if args.temp_prefix is None:
        args.temp_prefix = tempfile.gettempdir() + "/bcffilter"
    tp = Path(args.temp_prefix)
    if tp.is_dir() and not args.temp_prefix.endswith("/"):
        args.temp_prefix += "/"
    args.temp_prefix += str(uuid.uuid4())
    print(f"Using temporary file prefix: {args.temp_prefix}. This WILL NOT be automatically cleaned!!", file=stderr)

    chunkfiles = chunkwise_pipeline(args)
    merge_results(args, chunkfiles)

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import argparse

from Bio import SeqIO

def farename_main(argv=None):
    """Rename sequences in a fasta file sequentially"""
    ap = argparse.ArgumentParser()
    ap.add_argument("-p", "--prefix", type=str, default="",
            help="Prefix for each sequence ID (e.g. SEQ_ -> SEQ_1, SEQ_2 etc)")
    ap.add_argument("fastas", help="FASTA file of contigs", nargs="+")
    args = ap.parse_args(argv)
    
    for infile in args.fastas:
        i = 0
        with open(infile) as fh:
            for seq in SeqIO.parse(fh, 'fasta'):
                i += 1
                print(f">{args.prefix}{i} {seq.id} {seq.description}")
                print(seq.seq)

                
if __name__ == "__main__":
    farename_main()



#!/usr/bin/env python3
import re
import json
import select
from sys import stdin, stdout, stderr, exit
import sys

def main(argv=None):
    """INTERNAL: MPI Tübingen tool. Make a runlib-to-sample map table from ebio sra files"""
    samples = {}
    if not select.select([stdin,],[],[],2.0)[0]:
        print("This is an internal MPI Tübingen tool to make a table that mapps runs of a given library to sample names, e.g. for Acanthophis.", file=stderr)
        print(f"USAGE: find /ebio/abt6_sra/... -type f -name '*.fq.gz' | ebiosra2rl2s > run-lib2sample.tsv", file=stderr)
        exit(1)
    for line in stdin:
        m = re.match(r"(.*/illumina_.+_flowcell.+_SampleId(\w+)_(RunId\d+_LaneId\d+)/.+_(R\d)_\d+.fastq.gz)", line.strip())
        if m is not None:
            path, sample, run, read = m.groups()
            if sample not in samples:
                samples[sample] = {run: {read: path}}
            elif run not in samples[sample]:
                samples[sample][run] = {read: path}
            else:
                samples[sample][run][read] = path
    print("library", "run", "sample", "read1_uri", "read2_uri", file=stdout, sep='\t')
    for sample, dat in samples.items():
        for run, reads in dat.items():
            print(sample, run, sample, reads["R1"], reads["R2"], file=stdout, sep='\t')


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import argparse
from sys import stderr, stdout, stdin

KNOWN_HEADERS = {
        "##gff-version",
        "##feature-ontology",
}


def gffcat_main(argv=None):
    """Concatenate GFF3 files, resepcting header lines and FASTA sections"""
    ap = argparse.ArgumentParser("gffcat")
    ap.add_argument("-o", "--output", type=argparse.FileType("wt"), default=stdout,
            help="Output gff file")
    ap.add_argument("inputs", nargs="+")
    args = ap.parse_args(argv)

    fastalines = []
    bodylines = []
    headerlines = {}

    for file in args.inputs:
        with open(file) as fh:
            in_fasta = False
            for line in fh:
                line = line.rstrip()
                if not line:
                    continue
                ll = line.lower().split()[0]
                if ll in KNOWN_HEADERS:
                    if ll in headerlines:
                        if headerlines[ll] != line:
                            print("WARN: header line with different value in differnt files. Only using value from first file", file=stderr)
                    else:
                        headerlines[ll] = line
                elif ll.startswith("##fasta"):
                    in_fasta = True
                else:
                    if in_fasta:
                        fastalines.append(line)
                    else:
                        bodylines.append(line)

    for hdr in KNOWN_HEADERS:
        if hdr in headerlines:
            print(headerlines[hdr], file=args.output)

    for line in bodylines:
        print(line, file=args.output)

    if fastalines:
        print("##fasta", file=args.output)

    for line in fastalines:
        print(line, file=args.output)
    
    print(f"DONE! {len(headerlines)} headers, {len(bodylines)} body lines, {len(fastalines)} FASTA lines", file=stderr)


if __name__ == "__main__":
    gffcat_main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import argparse
import random
import time
import os
from tqdm import tqdm
try:
    import orjson as json
except ImportError:
    import json

def main(argv=None):
    """Parse jsonlines into a C/TSV"""
    ap = argparse.ArgumentParser()
    ap.add_argument("--fraction", "-f", type=float,
            help="Subsample this fraction of rows (0-1)")
    ap.add_argument("--seed", type=float, default=time.time() + os.getpid(),
            help="Seed the RNG with this seed (seeded adequately by default)")
    ap.add_argument("--header-same", "-s", action="store_true",
            help="Assume all rows have the same header, just use the first row's header")
    ap.add_argument("--out-csv", "-o", type=argparse.FileType("wt"), default="/dev/stdout",
            help="Output file")
    ap.add_argument("in_json")

    args=ap.parse_args(argv)
    header = set()
    first = True

    rand = random.Random(args.seed)
    with open(args.in_json) as fh:
        for line in tqdm(fh, desc="Read Headers", unit="lines"):
            if first or args.fraction is None or rand.random() < args.fraction:
                dat = json.loads(line)
                for key in dat:
                    header.add(key)
                if first and args.header_same:
                    break
                first = False

    try:
        sep = "\t" if args.out_csv.filename.endswith("tsv") else ","
    except:
        sep = ","
    rand = random.Random(args.seed)
    print(*header, sep=sep, file=args.out_csv)
    with open(args.in_json) as fh:
        for line in tqdm(fh, desc="Read Data", unit="lines"):
            if args.fraction is None or rand.random() < args.fraction:
                dat = json.loads(line)
                outline = [dat.get(key, "") for key in header]
                print(*outline, sep=sep, file=args.out_csv)
                

if __name__ == "__main__":
    main()
